@@ -1,11 +1,20 @@
 /*
- *   Copyright (c) 2003, 2004, 2005, 2006, 2007 Novell, Inc.
- *   (All rights reserved)
+ * (C) 2006, 2007 Andreas Gruenbacher <agruen@suse.de>
+ * Copyright (c) 2003-2008 Novell, Inc. (All rights reserved)
+ * Copyright 2009-2012 Canonical Ltd.
  *
- *   The libapparmor library is licensed under the terms of the GNU
- *   Lesser General Public License, version 2.1. Please see the file
- *   COPYING.LGPL.
- */
+ * The libapparmor library is licensed under the terms of the GNU
+ * Lesser General Public License, version 2.1. Please see the file
+ * COPYING.LGPL.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+  */
 
 #ifndef APPARMOR_RE_H
 #define APPARMOR_RE_H
@@ -17,10 +26,11 @@ typedef enum dfaflags {
   DFA_CONTROL_TREE_LEFT =	1 << 3,
   DFA_CONTROL_MINIMIZE =	1 << 4,
   DFA_CONTROL_MINIMIZE_HASH_TRANS = 1 << 5,
-  DFA_CONTROL_MINIMIZE_HASH_PERMS = 1 << 6,
+  DFA_CONTROL_FILTER_DENY =	1 << 6,
   DFA_CONTROL_REMOVE_UNREACHABLE =	1 << 7,
   DFA_CONTROL_TRANS_HIGH =	1 << 8,
 
+  DFA_DUMP_MIN_PARTS =		1 << 13,
   DFA_DUMP_UNIQ_PERMS =		1 << 14,
   DFA_DUMP_MIN_UNIQ_PERMS =	1 << 15,
   DFA_DUMP_TREE_STATS =		1 << 16,
@@ -41,25 +51,4 @@ typedef enum dfaflags {
   DFA_DUMP_NODE_TO_DFA =	1 << 31,
 } dfaflags_t;
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
-struct aare_ruleset;
-
-typedef struct aare_ruleset aare_ruleset_t;
-
-aare_ruleset_t *aare_new_ruleset(int reverse);
-void aare_delete_ruleset(aare_ruleset_t *rules);
-int aare_add_rule(aare_ruleset_t *rules, char *rule, int deny,
-		  uint32_t perms, uint32_t audit, dfaflags_t flags);
-int aare_add_rule_vec(aare_ruleset_t *rules, int deny, uint32_t perms,
-		      uint32_t audit, int count, char **rulev, dfaflags_t flags);
-void *aare_create_dfa(aare_ruleset_t *rules, size_t *size, dfaflags_t flags);
-void aare_reset_matchflags(void);
-
-#ifdef __cplusplus
-}
-#endif
-
 #endif /* APPARMOR_RE_H */
