@@ -1,7 +1,7 @@
 /*
  * (C) 2006, 2007 Andreas Gruenbacher <agruen@suse.de>
  * Copyright (c) 2003-2008 Novell, Inc. (All rights reserved)
- * Copyright 2009-2010 Canonical Ltd.
+ * Copyright 2009-2012 Canonical Ltd.
  *
  * The libapparmor library is licensed under the terms of the GNU
  * Lesser General Public License, version 2.1. Please see the file
@@ -30,11 +30,40 @@
 #include <ostream>
 #include <iostream>
 #include <fstream>
+#include <string.h>
 
 #include "expr-tree.h"
 #include "hfa.h"
 #include "../immunix.h"
 
+
+ostream &operator<<(ostream &os, const CacheStats &cache)
+{
+	/* dump the state label */
+	os << "cache: size=";
+	os << cache.size();
+	os << " dups=";
+	os << cache.dup;
+	os << " longest=";
+	os << cache.max;
+	if (cache.size()) {
+		os << " avg=";
+		os << cache.sum / cache.size();
+	}
+	return os;
+}
+
+ostream &operator<<(ostream &os, const ProtoState &proto)
+{
+	/* dump the state label */
+	os << '{';
+	os << proto.nnodes;
+	os << ',';
+	os << proto.anodes;
+	os << '}';
+	return os;
+}
+
 ostream &operator<<(ostream &os, const State &state)
 {
 	/* dump the state label */
@@ -44,46 +73,48 @@ ostream &operator<<(ostream &os, const State &state)
 	return os;
 }
 
-State *DFA::add_new_state(NodeMap &nodemap,
-			  pair<unsigned long, NodeSet *> index,
-			  NodeSet *nodes, dfa_stats_t &stats)
+static void split_node_types(NodeSet *nodes, NodeSet **anodes, NodeSet **nnodes
+)
 {
-	State *state = new State(nodemap.size(), nodes);
-	states.push_back(state);
-	nodemap.insert(make_pair(index, state));
-	stats.proto_sum += nodes->size();
-	if (nodes->size() > stats.proto_max)
-		stats.proto_max = nodes->size();
-	return state;
+	*anodes = *nnodes = NULL;
+	for (NodeSet::iterator i = nodes->begin(); i != nodes->end(); ) {
+		if ((*i)->is_accept()) {
+			if (!*anodes)
+				*anodes = new NodeSet;
+			(*anodes)->insert(*i);
+			NodeSet::iterator k = i++;
+			nodes->erase(k);
+		} else
+			i++;
+	}
+	*nnodes = nodes;
 }
 
-State *DFA::find_target_state(NodeMap &nodemap, list<State *> &work_queue,
-			      NodeSet *nodes, dfa_stats_t &stats)
+State *DFA::add_new_state(NodeSet *nodes, State *other)
 {
-	State *target;
+	/* The splitting of nodes should probably get pushed down into
+	 * follow(), ie. put in separate lists from the start
+	 */
+	NodeSet *anodes, *nnodes;
+	hashedNodeVec *nnodev;
+	split_node_types(nodes, &anodes, &nnodes);
+	nnodev = nnodes_cache.insert(nnodes);
+	anodes = anodes_cache.insert(anodes);
 
-	pair<unsigned long, NodeSet *> index = make_pair(hash_NodeSet(nodes), nodes);
-
-	map<pair<unsigned long, NodeSet *>, State *, deref_less_than>::iterator x = nodemap.find(index);
-
-	if (x == nodemap.end()) {
-		/* set of nodes isn't known so create new state, and nodes to
-		 * state mapping
-		 */
-		target = add_new_state(nodemap, index, nodes, stats);
-		work_queue.push_back(target);
+	ProtoState proto(nnodev, anodes);
+	State *state = new State(node_map.size(), proto, other);
+	pair<NodeMap::iterator,bool> x = node_map.insert(proto, state);
+	if (x.second == false) {
+		delete state;
 	} else {
-		/* set of nodes already has a mapping so free this one */
-		stats.duplicates++;
-		delete(nodes);
-		target = x->second;
+		states.push_back(state);
+		work_queue.push_back(state);
 	}
 
-	return target;
+	return x.first->second;
 }
 
-void DFA::update_state_transitions(NodeMap &nodemap, list<State *> &work_queue,
-				   State *state, dfa_stats_t &stats)
+void DFA::update_state_transitions(State *state)
 {
 	/* Compute possible transitions for state->nodes.  This is done by
 	 * iterating over all the nodes in state->nodes and combining the
@@ -91,9 +122,12 @@ void DFA::update_state_transitions(NodeMap &nodemap, list<State *> &work_queue,
 	 *
 	 * The resultant transition set is a mapping of characters to
 	 * sets of nodes.
+	 *
+	 * Note: the follow set for accept nodes is always empty so we don't
+	 * need to compute follow for the accept nodes in a protostate
 	 */
-	NodeCases cases;
-	for (NodeSet::iterator i = state->nodes->begin(); i != state->nodes->end(); i++)
+	Cases cases;
+	for (hashedNodeVec::iterator i = state->proto.nnodes->begin(); i != state->proto.nnodes->end(); i++)
 		(*i)->follow(cases);
 
 	/* Now for each set of nodes in the computed transitions, make
@@ -103,22 +137,22 @@ void DFA::update_state_transitions(NodeMap &nodemap, list<State *> &work_queue,
 
 	/* check the default transition first */
 	if (cases.otherwise)
-		state->cases.otherwise = find_target_state(nodemap, work_queue,
-							   cases.otherwise,
-							   stats);;
+		state->otherwise = add_new_state(cases.otherwise, nonmatching);
+	else
+		state->otherwise = nonmatching;
 
 	/* For each transition from *from, check if the set of nodes it
 	 * transitions to already has been mapped to a state
 	 */
-	for (NodeCases::iterator j = cases.begin(); j != cases.end(); j++) {
+	for (Cases::iterator j = cases.begin(); j != cases.end(); j++) {
 		State *target;
-		target = find_target_state(nodemap, work_queue, j->second, stats);
+		target = add_new_state(j->second, nonmatching);
 
-		/* Don't insert transition that the default transition
+		/* Don't insert transition that the otherwise transition
 		 * already covers
 		 */
-		if (target != state->cases.otherwise)
-			state->cases.cases[j->first] = target;
+		if (target != state->otherwise)
+			state->trans[j->first] = target;
 	}
 }
 
@@ -131,7 +165,7 @@ void DFA::dump_node_to_dfa(void)
 		"  State  <=   Nodes\n"
 		"-------------------\n";
 	for (Partition::iterator i = states.begin(); i != states.end(); i++)
-		cerr << "  " << (*i)->label << " <= " << *(*i)->nodes << "\n";
+		cerr << "  " << (*i)->label << " <= " << (*i)->proto << "\n";
 }
 
 /**
@@ -139,7 +173,6 @@ void DFA::dump_node_to_dfa(void)
  */
 DFA::DFA(Node *root, dfaflags_t flags): root(root)
 {
-	dfa_stats_t stats = { 0, 0, 0 };
 	int i = 0;
 
 	if (flags & DFA_DUMP_PROGRESS)
@@ -157,15 +190,8 @@ DFA::DFA(Node *root, dfaflags_t flags): root(root)
 		(*i)->compute_followpos();
 	}
 
-	NodeMap nodemap;
-	NodeSet *emptynode = new NodeSet;
-	nonmatching = add_new_state(nodemap,
-				    make_pair(hash_NodeSet(emptynode), emptynode),
-				    emptynode, stats);
-
-	NodeSet *first = new NodeSet(root->firstpos);
-	start = add_new_state(nodemap, make_pair(hash_NodeSet(first), first),
-			      first, stats);
+	nonmatching = add_new_state(new NodeSet, NULL);
+	start = add_new_state(new NodeSet(root->firstpos), nonmatching);
 
 	/* the work_queue contains the states that need to have their
 	 * transitions computed.  This could be done with a recursive
@@ -177,14 +203,18 @@ DFA::DFA(Node *root, dfaflags_t flags): root(root)
 	 *       manner, this may help reduce the number of entries on the
 	 *       work_queue at any given time, thus reducing peak memory use.
 	 */
-	list<State *> work_queue;
 	work_queue.push_back(start);
 
 	while (!work_queue.empty()) {
-		if (i % 1000 == 0 && (flags & DFA_DUMP_PROGRESS))
-			fprintf(stderr, "\033[2KCreating dfa: queue %zd\tstates %zd\teliminated duplicates %d\r",
-				work_queue.size(), states.size(),
-				stats.duplicates);
+		if (i % 1000 == 0 && (flags & DFA_DUMP_PROGRESS)) {
+			cerr << "\033[2KCreating dfa: queue "
+			     << work_queue.size()
+			     << "\tstates "
+			     << states.size()
+			     << "\teliminated duplicates "
+			     << node_map.dup
+			     << "\r";
+		}
 		i++;
 
 		State *from = work_queue.front();
@@ -193,7 +223,7 @@ DFA::DFA(Node *root, dfaflags_t flags): root(root)
 		/* Update 'from's transitions, and if it transitions to any
 		 * unknown State create it and add it to the work_queue
 		 */
-		update_state_transitions(nodemap, work_queue, from, stats);
+		update_state_transitions(from);
 
 	}  /* while (!work_queue.empty()) */
 
@@ -209,34 +239,68 @@ DFA::DFA(Node *root, dfaflags_t flags): root(root)
 	if (flags & DFA_DUMP_NODE_TO_DFA)
 		dump_node_to_dfa();
 
-	for (NodeMap::iterator i = nodemap.begin(); i != nodemap.end(); i++)
-		delete i->first.second;
-	nodemap.clear();
-
-	if (flags & (DFA_DUMP_STATS))
-		fprintf(stderr, "\033[2KCreated dfa: states %zd,\teliminated duplicates %d,\tprotostate sets: longest %u, avg %u\n",
-			states.size(), stats.duplicates, stats.proto_max,
-			(unsigned int)(stats.proto_sum / states.size()));
+	if (flags & (DFA_DUMP_STATS)) {
+		cerr << "\033[2KCreated dfa: states "
+		     << states.size()
+		     << " proto { "
+		     << node_map
+		     << " }, nnodes { "
+		     << nnodes_cache
+		     << " }, anodes { "
+		     << anodes_cache
+		     << " }\n";
+	}
 
+	/* Clear out uniq_nnodes as they are no longer needed.
+	 * Do not clear out uniq_anodes, as we need them for minimizations
+	 * diffs, unions, ...
+	 */
+	nnodes_cache.clear();
+	node_map.clear();
 }
 
 DFA::~DFA()
 {
+	anodes_cache.clear();
+	nnodes_cache.clear();
+
 	for (Partition::iterator i = states.begin(); i != states.end(); i++)
 		delete *i;
 }
 
+State *DFA::match_len(State *state, const char *str, size_t len)
+{
+	for (; len > 0; ++str, --len)
+		state = state->next(*str);
+
+	return state;
+}
+
+State *DFA::match_until(State *state, const char *str, const char term)
+{
+	while (*str != term)
+		state = state->next(*str++);
+
+	return state;
+}
+
+State *DFA::match(const char *str)
+{
+	return match_until(start, str, 0);
+}
+
 void DFA::dump_uniq_perms(const char *s)
 {
-	set<pair<uint32_t, uint32_t> > uniq;
+	set<perms_t> uniq;
 	for (Partition::iterator i = states.begin(); i != states.end(); i++)
-		uniq.insert(make_pair((*i)->accept, (*i)->audit));
+		uniq.insert((*i)->perms);
 
 	cerr << "Unique Permission sets: " << s << " (" << uniq.size() << ")\n";
 	cerr << "----------------------\n";
-	for (set<pair<uint32_t, uint32_t> >::iterator i = uniq.begin();
-	     i != uniq.end(); i++) {
-		cerr << "  " << hex << i->first << " " << i->second << dec << "\n";
+	for (set<perms_t >::iterator i = uniq.begin(); i != uniq.end(); i++) {
+		cerr << "  allow:" << hex << i->allow << " deny:"
+		     << i->deny << " audit:" << i->audit
+		     << " quiet:" << i->quiet << dec << "\n";
 	}
 }
 
@@ -244,7 +308,6 @@ void DFA::dump_uniq_perms(const char *s)
 void DFA::remove_unreachable(dfaflags_t flags)
 {
 	set<State *> reachable;
-	list<State *> work_queue;
 
 	/* find the set of reachable states */
 	reachable.insert(nonmatching);
@@ -254,11 +317,11 @@ void DFA::remove_unreachable(dfaflags_t flags)
 		work_queue.pop_front();
 		reachable.insert(from);
 
-		if (from->cases.otherwise &&
-		    (reachable.find(from->cases.otherwise) == reachable.end()))
-			work_queue.push_back(from->cases.otherwise);
+		if (from->otherwise != nonmatching &&
+		    reachable.find(from->otherwise) == reachable.end())
+			work_queue.push_back(from->otherwise);
 
-		for (Cases::iterator j = from->cases.begin(); j != from->cases.end(); j++) {
+		for (StateTrans::iterator j = from->trans.begin(); j != from->trans.end(); j++) {
 			if (reachable.find(j->second) == reachable.end())
 				work_queue.push_back(j->second);
 		}
@@ -277,12 +340,8 @@ void DFA::remove_unreachable(dfaflags_t flags)
 					cerr << "unreachable: " << **i;
 					if (*i == start)
 						cerr << " <==";
-					if ((*i)->accept) {
-						cerr << " (0x" << hex 
-						     << (*i)->accept << " " 
-						     << (*i)->audit << dec
-						     << ')';
-					}
+					if ((*i)->perms.is_accept())
+						(*i)->perms.dump(cerr);
 					cerr << "\n";
 				}
 				State *current = *i;
@@ -301,26 +360,17 @@ void DFA::remove_unreachable(dfaflags_t flags)
 /* test if two states have the same transitions under partition_map */
 bool DFA::same_mappings(State *s1, State *s2)
 {
-	if (s1->cases.otherwise && s1->cases.otherwise != nonmatching) {
-		if (!s2->cases.otherwise || s2->cases.otherwise == nonmatching)
-			return false;
-		Partition *p1 = s1->cases.otherwise->partition;
-		Partition *p2 = s2->cases.otherwise->partition;
-		if (p1 != p2)
-			return false;
-	} else if (s2->cases.otherwise && s2->cases.otherwise != nonmatching) {
+	if (s1->otherwise->partition != s2->otherwise->partition)
 		return false;
-	}
 
-	if (s1->cases.cases.size() != s2->cases.cases.size())
+	if (s1->trans.size() != s2->trans.size())
 		return false;
-	for (Cases::iterator j1 = s1->cases.begin(); j1 != s1->cases.end(); j1++) {
-		Cases::iterator j2 = s2->cases.cases.find(j1->first);
-		if (j2 == s2->cases.end())
+
+	for (StateTrans::iterator j1 = s1->trans.begin(); j1 != s1->trans.end(); j1++) {
+		StateTrans::iterator j2 = s2->trans.find(j1->first);
+		if (j2 == s2->trans.end())
 			return false;
-		Partition *p1 = j1->second->partition;
-		Partition *p2 = j2->second->partition;
-		if (p1 != p2)
+		if (j1->second->partition != j2->second->partition)
 			return false;
 	}
 
@@ -330,7 +380,7 @@ bool DFA::same_mappings(State *s1, State *s2)
 /* Do simple djb2 hashing against a States transition cases
  * this provides a rough initial guess at state equivalence as if a state
  * has a different number of transitions or has transitions on different
- * cases they will never be equivalent.
+ * trans they will never be equivalent.
  * Note: this only hashes based off of the alphabet (not destination)
  * as different destinations could end up being equiv
  */
@@ -338,22 +388,31 @@ size_t DFA::hash_trans(State *s)
 {
 	unsigned long hash = 5381;
 
-	for (Cases::iterator j = s->cases.begin(); j != s->cases.end(); j++) {
+	for (StateTrans::iterator j = s->trans.begin(); j != s->trans.end(); j++) {
 		hash = ((hash << 5) + hash) + j->first;
 		State *k = j->second;
-		hash = ((hash << 5) + hash) + k->cases.cases.size();
+		hash = ((hash << 5) + hash) + k->trans.size();
 	}
 
-	if (s->cases.otherwise && s->cases.otherwise != nonmatching) {
+	if (s->otherwise != nonmatching) {
 		hash = ((hash << 5) + hash) + 5381;
-		State *k = s->cases.otherwise;
-		hash = ((hash << 5) + hash) + k->cases.cases.size();
+		State *k = s->otherwise;
+		hash = ((hash << 5) + hash) + k->trans.size();
 	}
 
-	hash = (hash << 8) | s->cases.cases.size();
+	hash = (hash << 8) | s->trans.size();
 	return hash;
 }
 
+int DFA::apply_and_clear_deny(void)
+{
+	int c = 0;
+	for (Partition::iterator i = states.begin(); i != states.end(); i++)
+		c += (*i)->apply_and_clear_deny();
+
+	return c;
+}
+
 /* minimize the number of dfa states */
 void DFA::minimize(dfaflags_t flags)
 {
@@ -373,19 +432,11 @@ void DFA::minimize(dfaflags_t flags)
 	int accept_count = 0;
 	int final_accept = 0;
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		uint64_t perm_hash = 0;
-		if (flags & DFA_CONTROL_MINIMIZE_HASH_PERMS) {
-			/* make every unique perm create a new partition */
-			perm_hash = ((uint64_t) (*i)->audit) << 32 |
-				    (uint64_t) (*i)->accept;
-		} else if ((*i)->audit || (*i)->accept) {
-			/* combine all perms together into a single parition */
-			perm_hash = 1;
-		} /* else not an accept state so 0 for perm_hash */
-		size_t trans_hash = 0;
+		size_t hash = 0;
+		uint64_t permtype = ((uint64_t) (PACK_AUDIT_CTL((*i)->perms.audit, (*i)->perms.quiet & (*i)->perms.deny)) << 32) | (uint64_t) (*i)->perms.allow;
 		if (flags & DFA_CONTROL_MINIMIZE_HASH_TRANS)
-			trans_hash = hash_trans(*i);
-		pair<uint64_t, size_t> group = make_pair(perm_hash, trans_hash);
+			hash |= hash_trans(*i);
+		pair<uint64_t, size_t> group = make_pair(permtype, hash);
 		map<pair<uint64_t, size_t>, Partition *>::iterator p = perm_map.find(group);
 		if (p == perm_map.end()) {
 			Partition *part = new Partition();
@@ -393,7 +444,7 @@ void DFA::minimize(dfaflags_t flags)
 			perm_map.insert(make_pair(group, part));
 			partitions.push_back(part);
 			(*i)->partition = part;
-			if (perm_hash)
+			if (permtype)
 				accept_count++;
 		} else {
 			(*i)->partition = p->second;
@@ -487,11 +538,9 @@ void DFA::minimize(dfaflags_t flags)
 			cerr << *rep << " : ";
 
 		/* update representative state's transitions */
-		if (rep->cases.otherwise) {
-			Partition *partition = rep->cases.otherwise->partition;
-			rep->cases.otherwise = *partition->begin();
-		}
-		for (Cases::iterator c = rep->cases.begin(); c != rep->cases.end(); c++) {
+		rep->otherwise = *rep->otherwise->partition->begin();
+
+		for (StateTrans::iterator c = rep->trans.begin(); c != rep->trans.end(); c++) {
 			Partition *partition = c->second->partition;
 			c->second = *partition->begin();
 		}
@@ -505,10 +554,9 @@ void DFA::minimize(dfaflags_t flags)
 			if (flags & DFA_DUMP_MIN_PARTS)
 				cerr << **i << ", ";
 			(*i)->label = -1;
-			rep->accept |= (*i)->accept;
-			rep->audit |= (*i)->audit;
+			rep->perms.add((*i)->perms);
 		}
-		if (rep->accept || rep->audit)
+		if (rep->perms.is_accept())
 			final_accept++;
 //if ((*p)->size() > 1)
 //cerr << "\n";
@@ -563,26 +611,52 @@ out:
 void DFA::dump(ostream & os)
 {
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		if (*i == start || (*i)->accept) {
+		if (*i == start || (*i)->perms.is_accept()) {
 			os << **i;
 			if (*i == start)
-				os << " <==";
-			if ((*i)->accept) {
-				os << " (0x" << hex << (*i)->accept << " "
-				   << (*i)->audit << dec << ')';
-			}
+				os << " <== (allow/deny/audit/quiet)";
+			if ((*i)->perms.is_accept())
+				(*i)->perms.dump(os);
 			os << "\n";
 		}
 	}
 	os << "\n";
 
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		if ((*i)->cases.otherwise)
-			os << **i << " -> " << (*i)->cases.otherwise << "\n";
-		for (Cases::iterator j = (*i)->cases.begin();
-		     j != (*i)->cases.end(); j++) {
-			os << **i << " -> " << j->second << ":  "
-			   << j->first << "\n";
+		Chars excluded;
+
+		for (StateTrans::iterator j = (*i)->trans.begin();
+		     j != (*i)->trans.end(); j++) {
+			if (j->second == nonmatching) {
+				excluded.insert(j->first);
+			} else {
+				os << **i;
+				if ((*i)->perms.is_accept())
+					os << " ", (*i)->perms.dump(os);
+				os << " -> " << *(j)->second << ": 0x"
+				   << hex << (int) j->first;
+				if (isprint(j->first))
+					os << " " << j->first;
+				os << dec << "\n";
+			}
+		}
+
+		if ((*i)->otherwise != nonmatching) {
+			os << **i;
+			if ((*i)->perms.is_accept())
+				os << " ", (*i)->perms.dump(os);
+			os << " -> " << *(*i)->otherwise << ": [";
+			if (!excluded.empty()) {
+				os << "^";
+				for (Chars::iterator k = excluded.begin();
+				     k != excluded.end(); k++) {
+					if (isprint(*k))
+						os << *k;
+					else
+						os << "\\0x" << hex << (int) *k << dec;
+				}
+			}
+			os << "]\n";
 		}
 	}
 	os << "\n";
@@ -603,35 +677,42 @@ void DFA::dump_dot_graph(ostream & os)
 		if (*i == start) {
 			os << "\t\tstyle=bold" << "\n";
 		}
-		uint32_t perms = (*i)->accept;
-		if (perms) {
-			os << "\t\tlabel=\"" << **i << "\\n("
-			   << perms << ")\"" << "\n";
+		if ((*i)->perms.is_accept()) {
+			os << "\t\tlabel=\"" << **i << "\\n";
+			(*i)->perms.dump(os);
+			os << "\"\n";
 		}
 		os << "\t]" << "\n";
 	}
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		Cases &cases = (*i)->cases;
 		Chars excluded;
 
-		for (Cases::iterator j = cases.begin(); j != cases.end(); j++) {
+		for (StateTrans::iterator j = (*i)->trans.begin(); j != (*i)->trans.end(); j++) {
 			if (j->second == nonmatching)
 				excluded.insert(j->first);
 			else {
 				os << "\t\"" << **i << "\" -> \"" << *j->second
 				   << "\" [" << "\n";
-				os << "\t\tlabel=\"" << j->first << "\"\n";
-				os << "\t]" << "\n";
+				os << "\t\tlabel=\"";
+				if (isprint(j->first))
+					os << j->first;
+				else
+					os << "\\0x" << hex << (int) j->first << dec;
+
+				os << "\"\n\t]" << "\n";
 			}
 		}
-		if (cases.otherwise && cases.otherwise != nonmatching) {
-			os << "\t\"" << **i << "\" -> \"" << *cases.otherwise
+		if ((*i)->otherwise != nonmatching) {
+			os << "\t\"" << **i << "\" -> \"" << *(*i)->otherwise
 			   << "\" [" << "\n";
 			if (!excluded.empty()) {
 				os << "\t\tlabel=\"[^";
 				for (Chars::iterator i = excluded.begin();
 				     i != excluded.end(); i++) {
-					os << *i;
+					if (isprint(*i))
+						os << *i;
+					else
+						os << "\\0x" << hex << (int) *i << dec;
 				}
 				os << "]\"" << "\n";
 			}
@@ -651,11 +732,9 @@ map<uchar, uchar> DFA::equivalence_classes(dfaflags_t flags)
 	uchar next_class = 1;
 
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		Cases & cases = (*i)->cases;
-
 		/* Group edges to the same next state together */
 		map<const State *, Chars> node_sets;
-		for (Cases::iterator j = cases.begin(); j != cases.end(); j++)
+		for (StateTrans::iterator j = (*i)->trans.begin(); j != (*i)->trans.end(); j++)
 			node_sets[j->second].insert(j->first);
 
 		for (map<const State *, Chars>::iterator j = node_sets.begin();
@@ -742,10 +821,9 @@ void DFA::apply_equivalence_classes(map<uchar, uchar> &eq)
      */
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
 		map<uchar, State *> tmp;
-		tmp.swap((*i)->cases.cases);
-		for (Cases::iterator j = tmp.begin(); j != tmp.end(); j++)
-			(*i)->cases.cases.
-			    insert(make_pair(eq[j->first], j->second));
+		tmp.swap((*i)->trans);
+		for (StateTrans::iterator j = tmp.begin(); j != tmp.end(); j++)
+			(*i)->trans.insert(make_pair(eq[j->first], j->second));
 	}
 }
 
@@ -789,81 +867,66 @@ static inline int diff_qualifiers(uint32_t perm1, uint32_t perm2)
  * have any exact matches, then they override the execute and safe
  * execute flags.
  */
-uint32_t accept_perms(NodeSet *state, uint32_t *audit_ctl, int *error)
+int accept_perms(NodeSet *state, perms_t &perms)
 {
-	uint32_t perms = 0, exact_match_perms = 0;
-	uint32_t audit = 0, exact_audit = 0, quiet = 0, deny = 0;
+	int error = 0;
+	uint32_t exact_match_allow = 0;
+	uint32_t exact_audit = 0;
+
+	perms.clear();
+
+	if (!state)
+		return error;
 
-	if (error)
-		*error = 0;
 	for (NodeSet::iterator i = state->begin(); i != state->end(); i++) {
 		MatchFlag *match;
 		if (!(match = dynamic_cast<MatchFlag *>(*i)))
 			continue;
 		if (dynamic_cast<ExactMatchFlag *>(match)) {
 			/* exact match only ever happens with x */
-			if (!is_merged_x_consistent(exact_match_perms,
-						    match->flag) && error)
-				*error = 1;;
-			exact_match_perms |= match->flag;
+			if (!is_merged_x_consistent(exact_match_allow,
+						    match->flag))
+				error = 1;;
+			exact_match_allow |= match->flag;
 			exact_audit |= match->audit;
 		} else if (dynamic_cast<DenyMatchFlag *>(match)) {
-			deny |= match->flag;
-			quiet |= match->audit;
+			perms.deny |= match->flag;
+			perms.quiet |= match->audit;
 		} else {
-			if (!is_merged_x_consistent(perms, match->flag)
-			    && error)
-				*error = 1;
-			perms |= match->flag;
-			audit |= match->audit;
+			if (!is_merged_x_consistent(perms.allow, match->flag))
+				error = 1;
+			perms.allow |= match->flag;
+			perms.audit |= match->audit;
 		}
 	}
 
-//if (audit || quiet)
-//fprintf(stderr, "perms: 0x%x, audit: 0x%x exact: 0x%x eaud: 0x%x deny: 0x%x quiet: 0x%x\n", perms, audit, exact_match_perms, exact_audit, deny, quiet);
+	perms.allow |= exact_match_allow & ~(ALL_AA_EXEC_TYPE);
 
-	perms |= exact_match_perms & ~(AA_USER_EXEC_TYPE | AA_OTHER_EXEC_TYPE);
-
-	if (exact_match_perms & AA_USER_EXEC_TYPE) {
-		perms = (exact_match_perms & AA_USER_EXEC_TYPE) |
-			(perms & ~AA_USER_EXEC_TYPE);
-		audit = (exact_audit & AA_USER_EXEC_TYPE) |
-			(audit & ~AA_USER_EXEC_TYPE);
+	if (exact_match_allow & AA_USER_EXEC_TYPE) {
+		perms.allow = (exact_match_allow & AA_USER_EXEC_TYPE) |
+			(perms.allow & ~AA_USER_EXEC_TYPE);
+		perms.audit = (exact_audit & AA_USER_EXEC_TYPE) |
+			(perms.audit & ~AA_USER_EXEC_TYPE);
+		perms.exact = AA_USER_EXEC_TYPE;
 	}
-	if (exact_match_perms & AA_OTHER_EXEC_TYPE) {
-		perms = (exact_match_perms & AA_OTHER_EXEC_TYPE) |
-			(perms & ~AA_OTHER_EXEC_TYPE);
-		audit = (exact_audit & AA_OTHER_EXEC_TYPE) |
-			(audit & ~AA_OTHER_EXEC_TYPE);
+	if (exact_match_allow & AA_OTHER_EXEC_TYPE) {
+		perms.allow = (exact_match_allow & AA_OTHER_EXEC_TYPE) |
+			(perms.allow & ~AA_OTHER_EXEC_TYPE);
+		perms.audit = (exact_audit & AA_OTHER_EXEC_TYPE) |
+			(perms.audit & ~AA_OTHER_EXEC_TYPE);
+		perms.exact |= AA_OTHER_EXEC_TYPE;
 	}
-	if (perms & AA_USER_EXEC & deny)
-		perms &= ~AA_USER_EXEC_TYPE;
+	if (AA_USER_EXEC & perms.deny)
+		perms.deny |= AA_USER_EXEC_TYPE;
 
-	if (perms & AA_OTHER_EXEC & deny)
-		perms &= ~AA_OTHER_EXEC_TYPE;
+	if (AA_OTHER_EXEC & perms.deny)
+		perms.deny |= AA_OTHER_EXEC_TYPE;
 
-	perms &= ~deny;
+	perms.allow &= ~perms.deny;
+	perms.quiet &= perms.deny;
 
-	if (audit_ctl)
-		*audit_ctl = PACK_AUDIT_CTL(audit, quiet & deny);
-
-// if (perms & AA_ERROR_BIT) {
-//     fprintf(stderr, "error bit 0x%x\n", perms);
-//     exit(255);
-//}
-
-	//if (perms & AA_EXEC_BITS)
-	//fprintf(stderr, "accept perm: 0x%x\n", perms);
-	/*
-	   if (perms & ~AA_VALID_PERMS)
-	   yyerror(_("Internal error accumulated invalid perm 0x%llx\n"), perms);
-	 */
-
-//if (perms & AA_CHANGE_HAT)
-//     fprintf(stderr, "change_hat 0x%x\n", perms);
-
-	if (*error)
+	if (error)
 		fprintf(stderr, "profile has merged rule with conflicting x modifiers\n");
 
-	return perms;
+	return error;
 }
