@@ -1,5 +1,3 @@
-/* $Id$ */
-
 /*
  *   Copyright (c) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007
  *   NOVELL (All rights reserved)
@@ -61,6 +59,7 @@
 
 #define SUBDOMAIN_INTERFACE_VERSION 2
 #define SUBDOMAIN_INTERFACE_DFA_VERSION 5
+#define SUBDOMAIN_INTERFACE_POLICY_DB 16
 
 int sd_serialize_codomain(int option, struct codomain *cod);
 
@@ -77,7 +76,7 @@ static void print_error(int error)
 		PERROR(_("Out of memory\n"));
 		break;
 	case -EFAULT:
-		PERROR(_("Couldn't copy profile Bad memory address\n"));
+		PERROR(_("Couldn't copy profile: Bad memory address\n"));
 		break;
 	case -EPROTO:
 		PERROR(_("Profile doesn't conform to protocol\n"));
@@ -133,6 +132,9 @@ int load_codomain(int option, struct codomain *cod)
 			PERROR(_("%s: Unable to write to stdout\n"),
 			       progname);
 			break;
+		case OPTION_OFILE:
+			PERROR(_("%s: Unable to write to output file\n"),
+			       progname);
 		default:
 			PERROR(_("%s: ASSERT: Invalid option: %d\n"),
 			       progname, option);
@@ -155,6 +157,7 @@ int load_codomain(int option, struct codomain *cod)
 			       cod->name);
 			break;
 		case OPTION_STDOUT:
+		case OPTION_OFILE:
 			break;
 		default:
 			PERROR(_("%s: ASSERT: Invalid option: %d\n"),
@@ -451,57 +454,6 @@ inline int sd_write_listend(sd_serialize *p)
 	return 1;
 }
 
-int sd_serialize_pattern(sd_serialize *p, pcre *pat)
-{
-	if (!sd_write_struct(p, "pcre"))
-		return 0;
-	if (!sd_write32(p, pat->size - sizeof(pcre)))
-		return 0;
-	if (!sd_write32(p, pat->magic_number))
-		return 0;
-	if (!sd_write32(p, pat->options))
-		return 0;
-	if (!sd_write16(p, pat->top_bracket))
-		return 0;
-	if (!sd_write16(p, pat->top_backref))
-		return 0;
-	if (!sd_write8(p, pat->first_char))
-		return 0;
-	if (!sd_write8(p, pat->req_char))
-		return 0;
-	if (!sd_write8(p, pat->code[0]))
-		return 0;
-	if (!sd_write_blob(p, &pat->code[1], pat->size - sizeof(pcre), NULL))
-		return 0;
-	if (!sd_write_structend(p))
-		return 0;
-
-	return 1;
-}
-
-int sd_serialize_file_entry(sd_serialize *p, struct cod_entry *file_entry)
-{
-	PDEBUG("Writing file entry. name '%s'\n", file_entry->name);
-	if (!sd_write_struct(p, "fe"))
-		return 0;
-	if (!sd_write_string(p, file_entry->name, NULL))
-		return 0;
-	if (!sd_write32(p, file_entry->mode))
-		return 0;
-	if (!sd_write32(p, file_entry->pattern_type))
-		return 0;
-	if (file_entry->pattern_type == ePatternRegex) {
-		if (!sd_write_string(p, file_entry->pat.regex, NULL))
-			return 0;
-		if (!sd_serialize_pattern(p, file_entry->pat.compiled))
-			return 0;
-	}
-	if (!sd_write_structend(p))
-		return 0;
-
-	return 1;
-}
-
 int sd_serialize_dfa(sd_serialize *p, void *dfa, size_t size)
 {
 	if (dfa && !sd_write_aligned_blob(p, dfa, size, "aadfa"))
@@ -594,22 +546,9 @@ int count_tailglob_ents(struct cod_entry *list)
 	return count;
 }
 
-int count_pcre_ents(struct cod_entry *list)
-{
-	struct cod_entry *entry;
-	int count = 0;
-	list_for_each(list, entry) {
-		if (entry->pattern_type == ePatternRegex) {
-			count++;
-		}
-	}
-	return count;
-}
-
 int sd_serialize_profile(sd_serialize *p, struct codomain *profile,
 			 int flattened)
 {
-	struct cod_entry *entry;
 	uint64_t allowed_caps;
 
 	if (!sd_write_struct(p, "profile"))
@@ -671,15 +610,14 @@ int sd_serialize_profile(sd_serialize *p, struct codomain *profile,
 
 #define low_caps(X) ((u32) ((X) & 0xffffffff))
 #define high_caps(X) ((u32) (((X) >> 32) & 0xffffffff))
-	allowed_caps = (profile->capabilities | profile->set_caps) &
-		~profile->deny_caps;
+	allowed_caps = (profile->capabilities) & ~profile->deny_caps;
 	if (!sd_write32(p, low_caps(allowed_caps)))
 		return 0;
 	if (!sd_write32(p, low_caps(allowed_caps & profile->audit_caps)))
 		return 0;
 	if (!sd_write32(p, low_caps(profile->deny_caps & profile->quiet_caps)))
 		return 0;
-	if (!sd_write32(p, low_caps(profile->set_caps & ~profile->deny_caps)))
+	if (!sd_write32(p, 0))
 		return 0;
 
 	if (!sd_write_struct(p, "caps64"))
@@ -690,7 +628,7 @@ int sd_serialize_profile(sd_serialize *p, struct codomain *profile,
 		return 0;
 	if (!sd_write32(p, high_caps(profile->deny_caps & profile->quiet_caps)))
 		return 0;
-	if (!sd_write32(p, high_caps(profile->set_caps & ~profile->deny_caps)))
+	if (!sd_write32(p, 0))
 		return 0;
 	if (!sd_write_structend(p))
 		return 0;
@@ -717,6 +655,15 @@ int sd_serialize_profile(sd_serialize *p, struct codomain *profile,
 	} else if (profile->network_allowed)
 		pwarn(_("profile %s network rules not enforced\n"), profile->name);
 
+	if (profile->policy_dfa && regex_type == AARE_DFA) {
+		if (!sd_write_struct(p, "policydb"))
+			return 0;
+		if (!sd_serialize_dfa(p, profile->policy_dfa, profile->policy_dfa_size))
+			return 0;
+		if (!sd_write_structend(p))
+			return 0;
+	}
+
 	/* either have a single dfa or lists of different entry types */
 	if (regex_type == AARE_DFA) {
 		if (!sd_serialize_dfa(p, profile->dfa, profile->dfa_size))
@@ -725,47 +672,8 @@ int sd_serialize_profile(sd_serialize *p, struct codomain *profile,
 		if (!sd_serialize_xtable(p, profile->exec_table))
 			return 0;
 	} else {
-		/* pcre globbing entries */
-		if (count_pcre_ents(profile->entries)) {
-			if (!sd_write_list(p, "pgent"))
-				return 0;
-			list_for_each(profile->entries, entry) {
-				if (entry->pattern_type == ePatternRegex) {
-					if (!sd_serialize_file_entry(p, entry))
-						return 0;
-				}
-			}
-			if (!sd_write_listend(p))
-				return 0;
-		}
-
-		/* simple globbing entries */
-		if (count_tailglob_ents(profile->entries)) {
-			if (!sd_write_list(p, "sgent"))
-				return 0;
-			list_for_each(profile->entries, entry) {
-				if (entry->pattern_type == ePatternTailGlob) {
-					if (!sd_serialize_file_entry(p, entry))
-						return 0;
-				}
-			}
-			if (!sd_write_listend(p))
-				return 0;
-		}
-
-		/* basic file entries */
-		if (count_file_ents(profile->entries)) {
-			if (!sd_write_list(p, "fent"))
-				return 0;
-			list_for_each(profile->entries, entry) {
-				if (entry->pattern_type == ePatternBasic) {
-					if (!sd_serialize_file_entry(p, entry))
-						return 0;
-				}
-			}
-			if (!sd_write_listend(p))
-				return 0;
-		}
+		PERROR(_("Unknown pattern type\n"));
+		return 1;
 	}
 
 	if (profile->hat_table && regex_type != AARE_DFA) {
@@ -787,9 +695,13 @@ int sd_serialize_top_profile(sd_serialize *p, struct codomain *profile)
 {
 	int version;
 
-	if (regex_type == AARE_DFA)
-		version = SUBDOMAIN_INTERFACE_DFA_VERSION;
-	else
+	if (regex_type == AARE_DFA) {
+		/* Not yet
+		if (profile->policy_dfa)
+			version = SUBDOMAIN_INTERFACE_POLICYDB;
+		else */
+			version = SUBDOMAIN_INTERFACE_DFA_VERSION;
+	} else
 		version = SUBDOMAIN_INTERFACE_VERSION;
 
 
@@ -813,7 +725,7 @@ int sd_serialize_top_profile(sd_serialize *p, struct codomain *profile)
 int cache_fd = -1;
 int sd_serialize_codomain(int option, struct codomain *cod)
 {
-	int fd;
+	int fd = -1;
 	int error = -ENOMEM, size, wsize;
 	sd_serialize *work_area;
 	char *filename = NULL;
@@ -838,13 +750,16 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 		filename = "stdout";
 		fd = dup(1);
 		break;
+	case OPTION_OFILE:
+		fd = dup(fileno(ofile));
+		break;
 	default:
 		error = -EINVAL;
 		goto exit;
 		break;
 	}
 
-	if (fd < 0 && (kernel_load || option == OPTION_STDOUT)) {
+	if (fd < 0 && (kernel_load || option == OPTION_OFILE || option == OPTION_STDOUT)) {
 		PERROR(_("Unable to open %s - %s\n"), filename,
 		       strerror(errno));
 		error = -errno;
@@ -853,7 +768,7 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 
 	error = 0;
 
-	if (option != OPTION_STDOUT)
+	if (option != OPTION_STDOUT && option != OPTION_OFILE)
 		free(filename);
 
 	if (option == OPTION_REMOVE) {
@@ -861,10 +776,10 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 		int len = 0;
 
 		if (profile_namespace) {
-			len += strlen(profile_namespace) + 1;
+			len += strlen(profile_namespace) + 2;
 			ns = profile_namespace;
 		} else if (cod->namespace) {
-			len += strlen(cod->namespace) + 1;
+			len += strlen(cod->namespace) + 2;
 			ns = cod->namespace;
 		}
 		if (cod->parent) {
@@ -876,7 +791,7 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 				goto exit;
 			}
 			if (ns)
-				sprintf(name, "%s:%s//%s", ns,
+				sprintf(name, ":%s:%s//%s", ns,
 					cod->parent->name, cod->name);
 			else
 				sprintf(name, "%s//%s", cod->parent->name,
@@ -888,7 +803,7 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 				error = -errno;
 				goto exit;
 			}
-			sprintf(name, "%s:%s", ns, cod->name);
+			sprintf(name, ":%s:%s", ns, cod->name);
 		} else {
 			name = cod->name;
 		}
@@ -919,7 +834,7 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 		}
 
 		size = work_area->pos - work_area->buffer;
-		if (kernel_load || option == OPTION_STDOUT) {
+		if (kernel_load || option == OPTION_STDOUT || option == OPTION_OFILE) {
 			wsize = write(fd, work_area->buffer, size);
 			if (wsize < 0) {
 				error = -errno;
@@ -942,7 +857,7 @@ int sd_serialize_codomain(int option, struct codomain *cod)
 		free_sd_serial(work_area);
 	}
 
-	if (kernel_load) close(fd);
+	close(fd);
 
 	if (cod->hat_table && regex_type == AARE_DFA && option != OPTION_REMOVE) {
 		if (load_flattened_hats(cod) != 0)
@@ -974,7 +889,7 @@ static char *next_profile_buffer(char *buffer, int size)
 
 int sd_load_buffer(int option, char *buffer, int size)
 {
-	int fd;
+	int fd = -1;
 	int error = -ENOMEM, wsize, bsize;
 	char *filename = NULL;
 	char *b;
